# Copyright (C) 2020 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from collections.abc import Iterable
from typing import Any

from stevedore.example import base


class Simple(base.FormatterBase):
    """A very basic formatter."""

    def format(self, data: dict[str, Any]) -> Iterable[str]:
        """Format the data and return unicode text.

        :param data: A dictionary with string keys and simple types as values.
        :returns: Iterable producing the formatted text.
        """
        for name, value in sorted(data.items()):
            line = f'{name} = {value}\n'
            yield line
